//***********************************************************************//
//*        FASTMAP.CPP --- Implementation of Fastmap algorithm          *//
//*Reference:"Fast Map: A Fast Algorithm for Indexing, Data-Mining and  *//
//*Visualization of Traditional and Multimedia Datasets"                *//
//*                                                                     *//
//*Copyright (C) Liang Jin (liangj@ics.uci.edu)                         *//
//*              Chen Li (chenli@ics.uci.edu)                           *//
//*Univ of California,Irvine Department of Information&Computer Science *//
//*                                                                     *//
//*This program is free software; you can redistribute itand/or         *//
//*modify it under the terms of the GNU General Public License          *//
//*as published by the Free Software Foundation; either version 2       *//
//*of the License, or (at your option) any later version.               *//
//*                                                                     *//
//*This program is distributed in the hope that it will be useful,      *//
//*but WITHOUT ANY WARRANTY; without even the implied warranty of       *//
//*MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *//
//*GNU General Public License for more details.                         *//
//*                                                                     *//
//*You should have received a copy of the GNU General Public License    *//
//*along with this program; if not, write to the Free Software          *//
//*Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            *//
//*02111-1307, USA.                                                     *//
//*                                                                     *//
//*                                                                     *//
//*This code was licensed under the GNU public license. For our         *//
//*modifications,the Regents of the University of California demand the *//
//*following terms: (where a conflic occurs, the lesser restrictive term*//
//*applies)                                                             *//
//*                                                                     *//
//*Copyright 1996-99 The Regents of the University of California.       *//
//*All Rights Reserved.                                                 *//
//*Permission to use, copy, modify, and distribute this software        *//
//*and its documentation for educational, research and non-profit       *//
//*purposes, without fee, and without a written agreement is hereby     *//
//*granted, provided that the above copyright notice, this paragraph    *//
//*and the following three paragraphs appear in all copies.             *//
//*Permission to incorporate this software into commercial products     *//
//*may be obtained by contacting the University of California.          *//
//*Chen Li Department of Information and Computer Science (ICS)         *//
//*University of California Irvine, California 92697-3425               *//
//*Phone: 949-824-9470.                                                 *//
//*                                                                     *//
//*This software program and documentation are copyrighted by           *//
//*The Regents of the University of California. The software program    *//
//*and documentation are supplied "as is", without any accompanying     *//
//*services from The Regents. The Regents do not warrant that the       *//
//*operation of the program will be uninterrupted or error-free.        *//
//*The end-user understands that the program was developed for          *//
//*research purposes and is advised not to rely exclusively on          *//
//*the program for any reason.                                          *//
//*                                                                     *//
//*IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO          *//
//*ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR              *//
//*CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT           *//
//*OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE       *//
//*UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF      *//
//*SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS     *//
//*ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           *//
//*WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  *//
//*THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE      *//
//*UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,  *//
//*SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                    *//
//***********************************************************************//





#include <fstream>
#include <iostream>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "fastmap.h"
#include "parameters.h"
#include "distance.h"
#include "MyPoint.h"
using namespace std;

#define getloc(i,j)	( *(loc+i*NUMDIMS+j) )

Fastmap::Fastmap(char **name)
{
  	names = name;
}

void Fastmap::run()
{
	srand((unsigned)time(NULL));

	loc=(double*)malloc(SIZE*NUMDIMS*sizeof(double));


	int i=0,j=0;

   	col=0;
	nextfm(NUMDIMS);

	cout<<"finish recursive call" << endl;

	ofstream location;
	location.open(DATASET1_SRC);

	for(i=0;i<SIZE/2;i++)
	{
		if(i==0)
			location << "1000000" ;
		else
			location << i;

		for(j=0;j<NUMDIMS;j++)
		{
			location << " " << getloc(i,j);

		}

		for(j=0;j<NUMDIMS;j++)
		{
			location << " " << getloc(i,j);
		}

		location << endl;
	}
	location.close();

	location.open(DATASET2_SRC);

	for(i=SIZE/2;i<SIZE;i++)
	{
		location << i;

		for(j=0;j<NUMDIMS;j++)
		{
			location << " " << getloc(i,j);

		}

		for(j=0;j<NUMDIMS;j++)
		{
			location << " " << getloc(i,j);
		}

		location << endl;
	}
	location.close();

	free(loc);
}

void Fastmap::nextfm(int kd)
{
	if(kd<=0)
		return;
	MyPoint point;

	//get the pivot points and record them
	choosepivot(&point);
	PA[col][0]=point.x;
	PA[col][1]=point.y;


	double xy;
	xy=getDistance(point.x,point.y,col);
	if(xy==0)
	{
		for(int k=0;k<SIZE;k++)
			getloc(k,col)=0;

		col++;
		nextfm(kd-1);
	}

	//set the value for each point in one NUMDIMS
	double xi,xk,yk;
	for(int k=0;k<SIZE;k++)
	{
		xk=getDistance(k,point.x,col);
		yk=getDistance(k,point.y,col);

		xi=(xk*xk+xy*xy-yk*yk)/(2*xy);
		getloc(k,col)=xi;
	}

	col++;
	nextfm(kd-1);
}

void Fastmap::choosepivot(MyPoint *pivot)
{

	int j=rand();
	double temp=j%SIZE;
	int seedb=(int)ceil(temp);
	int seeda;
	int farthest;
	double longestdis;
	int counter;
	double dis;
	for(int k=0;k<5;k++) //repeat the heuristic 5 times
	{
		farthest=0;
		longestdis=0;
		for(counter=0;counter<SIZE;counter++)
		{/*distance[counter][seedb]*/
			dis=getDistance(counter,seedb,col);
			if(dis>longestdis)
			{
				farthest=counter;
				longestdis=dis;
			}			/*distance[counter][seedb]*/
		}
		seeda=farthest;
		farthest=0;
		longestdis=0;
		for(counter=0;counter<SIZE;counter++)
		{/*distance[counter][seeda]*/
			dis=getDistance(counter,seeda,col);
			if(dis>longestdis)
			{
				farthest=counter;
				longestdis=dis;
			}			/*distance[counter][seeda]*/
		}
		seedb=farthest;
	}

	pivot->x=seeda;
	pivot->y=seedb;
	return;

}


double Fastmap::getDistance(int i, int j, int k)
{
	Distance1 dis1;
	double editdis,realdis,tmp;
	editdis=dis1.LD(names[i],names[j]);

	realdis=editdis;

	//the first iteration, just return the edit distance
	if(k==0)
		return editdis;

	for(int p=0;p<k;p++)
	{
		if(getloc(i,p)>getloc(j,p))
			tmp=getloc(i,p)-getloc(j,p);
		else
			tmp=getloc(j,p)-getloc(i,p);

		if(tmp>realdis)
			realdis=sqrt(tmp*tmp-realdis*realdis);
		else
			realdis=sqrt(realdis*realdis-tmp*tmp);
	}

	return realdis;
}
